import { defineStore } from "pinia";
import { useConnectionStore } from "./connection";
import * as Y from "yjs";
import * as AwarenessProtocol from "y-protocols/awareness.js";

export const useMissionDetailStore = defineStore("missionDetail", {
  state: () => {
    return {
      yDoc: new Y.Doc(),
      awareness: undefined as undefined | AwarenessProtocol.Awareness,
      docId: null as null | string,
      lastUpdateTimestamp: 0 as number,
      connectionStatus: "disconnected", // 'disconnected', 'connecting', 'connected', 'syncing', 'synced'
    };
  },
  actions: {
    initialize(docId: string) {
      this.docId = docId;

      this.lastUpdateTimestamp = this.loadLastUpdateFromLocalStorage();
      this.awareness = new AwarenessProtocol.Awareness(this.yDoc);
      this.awareness.setLocalStateField("user", { name: "hi", color: "#123456" });

      this.setupSocketHandlers();
      this.setupYjsObservers();
      this.setupAwarenessObservers();
    },

    setupSocketHandlers() {
      const connectionStore = useConnectionStore();

      connectionStore.connection?.on("package-sync", (data) => {
        try {
          this.connectionStatus = "syncing";

          Y.applyUpdate(this.yDoc, new Uint8Array(data.update));

          if (data.timestamp > this.lastUpdateTimestamp) {
            this.lastUpdateTimestamp = data.timestamp;
            this.saveLastUpdateToLocalStorage();
          }

          this.connectionStatus = "synced";
        } catch (error) {
          console.error("Error applying update:", error);
          this.requestFullSync();
        }
      });
      connectionStore.connection?.on("sync-get-missing-updates", (data) => {
        const clientUpdates = Y.encodeStateAsUpdate(this.yDoc, new Uint8Array(data.stateVector));

        connectionStore.connection?.emit("mission:sync-client-updates", {
          update: clientUpdates,
          timestamp: Date.now(),
        });
      });

      connectionStore.connection?.on("package-sync-awareness", (data) => {
        // if (this.awareness != undefined) {
        //   AwarenessProtocol.applyAwarenessUpdate(this.awareness, new Uint8Array(data.update), this);
        // }
      });

      this.joinDocument();
    },

    setupYjsObservers() {
      this.yDoc.on("update", (update) => {
        const connectionStore = useConnectionStore();
        if (connectionStore.connected) {
          connectionStore.connection?.emit("mission:sync-client-updates", {
            update: Array.from(update),
            timestamp: Date.now(),
          });
        }

        this.lastUpdateTimestamp = Date.now();
        this.saveLastUpdateToLocalStorage();
      });
    },

    setupAwarenessObservers() {
      if (!this.awareness) return;

      this.awareness.on("update", (update: { added: number[]; updated: number[]; removed: number[] }) => {
        if (this.awareness != undefined) {
          const changedClients = update.added.concat(update.updated).concat(update.removed);
          const connectionStore = useConnectionStore();
          connectionStore.connection?.emit(
            "mission:sync-client-awareness",
            Array.from(AwarenessProtocol.encodeAwarenessUpdate(this.awareness, changedClients))
          );
        }
      });
    },

    joinDocument() {
      if (!this.docId) return;

      const connectionStore = useConnectionStore();
      connectionStore.connection?.emit("mission:join", this.docId, {
        timestamp: this.lastUpdateTimestamp,
      });
    },

    requestFullSync() {
      if (!this.docId) return;

      const connectionStore = useConnectionStore();
      connectionStore.connection?.emit("mission:join", this.docId, null);
    },

    loadLastUpdateFromLocalStorage() {
      if (!this.docId) return 0;

      const stored = localStorage.getItem(`yjsDoc_timestamp`);
      return stored ? parseInt(stored, 10) : 0;
    },

    saveLastUpdateToLocalStorage() {
      localStorage.setItem(`yjsDoc_timestamp`, this.lastUpdateTimestamp.toString());
    },

    cleanup() {
      if (this.awareness) {
        AwarenessProtocol.removeAwarenessStates(this.awareness, [this.yDoc.clientID], "window unload");
        this.awareness.destroy();
      }

      this.yDoc.destroy();
      this.yDoc = new Y.Doc();

      this.lastUpdateTimestamp = 0;
      localStorage.removeItem("yjsDoc_timestamp");

      const connectionStore = useConnectionStore();
      connectionStore.connection?.emit("mission:leave");

      this.connectionStatus = "disconnected";
    },
  },
});
